//
//
// DoCMStuff.cpp
//
//


extern "C" {
#include <windows.h>
#include "vmm.h"

#define Not_VxD
#define IS_32
#define CM_USE_OPEN_SERVICE

#include "configmg.h"
#include <stdio.h>      // for sprintf
}

CM_IS_FILE_PROVIDING_SERVICES

// strings used in this file...
const char STR_HARDWAREID[] = "HardwareID";
const char STR_PNPB02F[] = "*PNPB02F";
const char STR_COMPATIBLEIDS[] = "CompatibleIDs";

// working gameport of a Crystal Soundcard:
// entry: Crystal PNP Audio System Joystick
// characeteristics: HardwareID = "*CSC0001"
const char STR_CSC0001[] = "*CSC0001";

// a gameport for the PCI soundcard YAMAHA DS-XG
// entry: "YAMAHA DS-XG GamePort"
// characeteristics: HardwareID = "YDSXG\\*YMH0101,*YMH0101"
const char STR_YMH0101[] = "*YMH0101";


void IsGamePort(DEVNODE devInst, int &nGPNum, int &nGPOK)
{
    BOOL bItIs = FALSE;
    CONFIGRET cr = CR_SUCCESS;
    
    ULONG ulStatus = 0;
    ULONG ulProblemNumber = 0;
    ULONG ulRegValLen = 0;
    char *pRegValue;

    /*
    char deviceid[256];
    ULONG size = 0;
    CM_Get_Device_ID_Size(&size, devInst, 0);
    if (256 > (size+1))
    {
        cr = CM_Get_Device_ID(devInst, deviceid, size+1, 0);
        OutputDebugString(deviceid);
        OutputDebugString("\n");
    }
    */

    cr = CM_Read_Registry_Value(devInst, NULL, (char *) STR_HARDWAREID, REG_SZ, NULL, &ulRegValLen, 0);
    if(CR_SUCCESS == cr)
    {
        pRegValue = new char[ulRegValLen+4];
        
        cr = CM_Read_Registry_Value(devInst, NULL, (char *) STR_HARDWAREID, REG_SZ, pRegValue, &ulRegValLen, 0);
        if(CR_SUCCESS == cr)
        {
            bItIs = (NULL != strstr(pRegValue, STR_PNPB02F));
            bItIs = bItIs || (NULL != strstr(pRegValue, STR_CSC0001));
            bItIs = bItIs || (NULL != strstr(pRegValue, STR_YMH0101));
        }
        delete pRegValue;
    }
    
    cr = CM_Read_Registry_Value(devInst, NULL, (char *) STR_COMPATIBLEIDS, REG_SZ, NULL, &ulRegValLen, 0);
    if(CR_SUCCESS == cr)
    {
        pRegValue = new char[ulRegValLen+4];
        
        cr = CM_Read_Registry_Value(devInst, NULL, (char *) STR_COMPATIBLEIDS, REG_SZ, pRegValue, &ulRegValLen, 0);
        if(CR_SUCCESS == cr)
        {
            bItIs = bItIs || (NULL != strstr(pRegValue, STR_PNPB02F));
            bItIs = bItIs || (NULL != strstr(pRegValue, STR_CSC0001));
            bItIs = bItIs || (NULL != strstr(pRegValue, STR_YMH0101));
        }
        delete pRegValue;
    }
    
    if(bItIs)
        nGPNum++;
    
    cr = CM_Get_DevNode_Status(&ulStatus, &ulProblemNumber, devInst, 0);
    bItIs = bItIs && (CR_SUCCESS == cr) && (0 == ulProblemNumber);
    
    if(bItIs)
        nGPOK++;
}


void EnumDevs(DEVNODE devInst, int &nGPNum, int &nGPOK)
{
    CONFIGRET cr = CR_SUCCESS;
    DEVNODE devChild = 0;
    DEVNODE devSibling = 0;
    
    if(0 == devInst)
        cr = CM_Locate_DevNode(&devInst, NULL, 0);

    if(cr == CR_SUCCESS)
    {
        IsGamePort(devInst, nGPNum, nGPOK);
        
        if((cr = CM_Get_Child(&devChild, devInst, 0)) == CR_SUCCESS)
            EnumDevs(devChild, nGPNum, nGPOK);
        if((cr = CM_Get_Sibling(&devSibling, devInst, 0)) == CR_SUCCESS)
            EnumDevs(devSibling, nGPNum, nGPOK);
    }
}

BOOL FindDevNode(DEVNODE dnHere, PDEVNODE dnResult, PFARCHAR partialName)
{
    if(0 == dnHere)
        return FALSE;

    CONFIGRET cr = CR_SUCCESS;
    DEVNODE dnChild = 0;
    DEVNODE dnSibling = 0;
    char deviceid[256];
    
    CM_Get_Device_ID(dnHere, deviceid, sizeof(deviceid), 0);
    if((CR_SUCCESS == cr) && (0 == strnicmp(deviceid, partialName, strlen(partialName))))
    {
        *dnResult = dnHere;
        return TRUE;
    }
    
    if(CR_SUCCESS == (cr = CM_Get_Child(&dnChild, dnHere, 0)))
        if(FindDevNode(dnChild, dnResult, partialName))
            return TRUE;

    if(CR_SUCCESS == (cr = CM_Get_Sibling(&dnSibling, dnHere, 0)))
        if(FindDevNode(dnSibling, dnResult, partialName))
            return TRUE;
    
    return FALSE;
}


void RemoveDeviceNevertheless(const char *deviceID)
{
    if(CM_Open())
    {
        CM_Lock(0);

        DEVNODE dnUSB;
        DWORD cr = CM_Locate_DevNode(&dnUSB, NULL, 0);
        if((CR_SUCCESS == cr) && FindDevNode(dnUSB, &dnUSB, (PFARCHAR) deviceID))
        {
            if(CR_SUCCESS == (cr = CM_Query_Remove_SubTree(dnUSB, 0)))
            {
                if(CR_SUCCESS != (cr = CM_Remove_SubTree(dnUSB, 0)))
                {
                    OutputDebugString("Problem removing..\n");
                }
                if(CR_SUCCESS != (cr = CM_Process_Events_Now(0)))
                {
                    OutputDebugString("Problem processing events...\n");
                }
            }
        }
        
        CM_Unlock(0);

        CM_Close();
    }
}


BOOL RemoveDeviceIfProblematic(const char *deviceID)
{
    BOOL bHasProblems = FALSE;

    if(CM_Open())
    {
        DEVNODE dnUSB;
        DWORD cr = CM_Locate_DevNode(&dnUSB, NULL, 0);
        if((CR_SUCCESS == cr) && FindDevNode(dnUSB, &dnUSB, (PFARCHAR) deviceID))
        {
            ULONG ulStatus, ulProblem;
            DWORD cr = CM_Get_DevNode_Status(&ulStatus, &ulProblem, dnUSB, 0);
            bHasProblems = (0 != ulProblem);
            if(bHasProblems)
            {
                // this node has a problem... let's remove it...
                char keybuff[1024];
                cr = CM_Get_DevNode_Key(dnUSB, NULL, keybuff, sizeof(keybuff), CM_REGISTRY_HARDWARE);
                if(CR_SUCCESS == (cr = CM_Query_Remove_SubTree(dnUSB, 0)))
                {
                    if(CR_SUCCESS != (cr = CM_Remove_SubTree(dnUSB, 0)))
                    {
                        OutputDebugString("Problem removing failed USB stuff..\n");
                    }
                    // remove registry entries...
                    LONG lResult = ERROR_SUCCESS;
                    char *p = keybuff;
                    while(strstr(keybuff, deviceID) && (ERROR_SUCCESS == lResult))
                    {
                        lResult = RegDeleteKey(HKEY_LOCAL_MACHINE, keybuff);
                        char *p = strrchr(keybuff, '\\');
                        if(p)
                            *p = 0x00;
                    }
                }
            }
        }
        CM_Close();
    }

    return bHasProblems;
}


void Reenumerate(void)
{
    if(CM_Open())
    {
        OutputDebugString("Locking CM...\n");
        CM_Lock(0);

        DEVNODE dn = NULL;
        if(CR_SUCCESS == CM_Locate_DevNode(&dn, NULL, 0))
        {
            OutputDebugString("About to do ReEnumerateDevNode()...\n");
            CM_Reenumerate_DevNode(dn, CM_REENUMERATE_SYNCHRONOUS);
        }

        OutputDebugString("About to do ProcessEventsNow()...\n");
        if(CR_SUCCESS != CM_Process_Events_Now(0))
        {
            OutputDebugString("Problem while doing events now...\n");
        }

        OutputDebugString("Unlocking CM...\n");
        CM_Unlock(0);

        OutputDebugString("Closing CM...\n");
        CM_Close();
    }
}


void IsGamePortInstalled(int &nGPNum, int &nGPOK)
{
    if(CM_Open())
    {
        EnumDevs(0, nGPNum, nGPOK);
        CM_Close();
    }
}

BOOL InstallGamePort(DWORD &err)
{
// to install we create this node in the registry and let
// the configuration manager reenumerate the root :-)
// [HKEY_LOCAL_MACHINE\Enum\Root\*PNPB02F\0000]
// "HardwareID"="*PNPB02F"

#ifndef LWTEST
    err = WinExec("RunDll.EXE Sysdm.cpl,InstallDevice_RunDll Media", SW_SHOWNORMAL);
    return (31 < err);
#else

    BOOL bResult = FALSE;
    const TCHAR DEVICEID[] = "Root\\*PNPB02F\\0000";
    DEVNODE devInst = NULL;
    err = CR_FAILURE;

    if(CM_Open())
    {
        err = CM_Locate_DevNode(&devInst, NULL, 0);
        if(CR_SUCCESS == err)
        {
            err = CM_Create_DevNode(NULL, "Root\\*PNPB02F\\0000", devInst, 0);
            //cr = CM_Write_Registry_Value(devInst, "Root\\*PNPB02F\\0000", "HardwareID", REG_SZ, "*PNPB02F", sizeof(PNPID), 0);
            //cr = CM_Reenumerate_DevNode(devInst, 0);
        }
        CM_Close();
    }

    return (CR_SUCCESS == err);
#endif
}
